module ExprParser where

import Prelude hiding (div)
import Char
import Language.HaLex.Parser

import Types


{-
parsing the following grammar:

Exp -> Factor AddOp Exp
    |  Factor

AddOp -> '+'
      |  '*'

Factor -> Termo MulOp Factor
       |  Termo

MulOp -> '*'
      |  '/'

Termo -> Real
      |  Ref

Real -> int
      | int '.' int

Ref -> 'c(' int ',' int ')'

int = (0..9)+

After left factorization (for efficiency purposes), this grammar remains the same, except for
Exp and Factor productions, which become:

Exp    -> Factor (AddOp Exp | empty)

Factor -> Termo (MulOp Factor | empty)

-}

zeroOuMais parser =  succeed [] 
		 <|> umOuMais parser

umOuMais parser = (:[]) <$> parser 
	       <|> (:)  <$> parser <*> umOuMais parser

parseAddOp =  plus <$> symbol '+'
	  <|> sub  <$> symbol '-'
	where plus = id
	      sub  = id

parseMulOp =  mul <$> symbol '*'
	  <|> div <$> symbol '/'
	where mul = id
	      div = id

parseReal :: [Char] -> [(Float, [Char])]
parseReal =  int   <$> umOuMais (satisfy isDigit)
	 <|> float <$> umOuMais (satisfy isDigit) <*> symbol '.' <*> umOuMais (satisfy isDigit)
	where int   x     = (read x) :: Float
	      float a _ b = (fromIntegral (read a))+((fromIntegral (read b))/10*fromIntegral(length b)) 



parseTermo env (r,c) =  real (r,c) <$> parseReal
	  	    <|> ref  (r,c) <$> parseRef  env
	where real (r,c) v = Const (r,c) v
	      ref  (r,c) v = Cell  (r,c) v

parseRef env = cel env <$> token "c(" <*> umOuMais (satisfy isDigit) <*> symbol ',' <*> umOuMais (satisfy isDigit) <*> symbol ')'
	where cel env _ r' _ c' _ = 
		case lookup (read r',read c') env of --Nothing                -> []
					             Just (_,value) -> CR (read r',read c') value

parseFactor env (r,c) = factor <$> parseTermo env (r,c) <*> (expr <$> parseMulOp <*> parseExp env (r,c)
		      					     <|> succeed [])
	where expr op e               = [(op,e)]
	      factor term [('*',exp)] = Mul (r,c) [term,exp] 
	      factor term [('/',exp)] = Div (r,c) [term,exp] 
	      factor term []           = term 

parseExp ::    [((Int, Int), (IOMod (GnuExpr GnuCellRef), IOMod Float))]
            -> (Int, Int) 
	    -> [Char] 
            -> [(GnuExpr GnuCellRef, [Char])]
parseExp env (r,c) = factor <$> parseFactor env (r,c) <*> (expr <$> parseAddOp <*> parseExp env (r,c)
							   <|> succeed [])
	where expr op e = [(op,e)]
	      factor fact [('+',exp)] = Add (r,c) [fact,exp] 
	      factor fact [('-',exp)] = Sub (r,c) [fact,exp]
	      factor fact []          = fact 


