module Main where

import Prelude hiding (subtract)

import Language.Gnumeric.Syntax
import Language.Gnumeric.SyntaxTermInstances
import Language.Gnumeric.Formula
import Language.Gnumeric.FormulaTermInstances
import Language.Gnumeric.Parser

import Text.XML.HaXml.Xml2Haskell
import Text.XML.HaXml.Types

import Adaptive

import Monad 
import Maybe
import IOExts(IORef)
import System
import Char

import Types
import ExprParser
import Operations


intoAdaptive :: (NewMod n IO IORef) =>
                [Gmr'Cell] -> [n IO IORef (GnuExpr GnuCellRef)]
intoAdaptive []  = []
intoAdaptive ((Gmr'Cell attrs str):cs) = 
	let pos  = (read (gmr'CellRow attrs), read (gmr'CellCol attrs))
	    expr = gmr'CellAst attrs
	in (intoAdaptive' pos expr):(intoAdaptive cs) 

intoAdaptive' :: (NewMod n IO IORef) =>
                     (Int, Int)
                       -> Maybe Formula
                          -> (n IO IORef (GnuExpr GnuCellRef))
intoAdaptive' pos (Just (Var x))    = let val = read x
				      in  return (Const pos val)
					       
intoAdaptive' pos (Just (Fun op l)) = let aux = sequence $ map ((intoAdaptive' pos).Just) l
				      in case op of "+" -> aux >>= return.(Add pos)
				   	 	    "*" -> aux >>= return.(Mul pos)
						    "-" -> aux >>= return.(Sub pos)
						    "/" -> aux >>= return.(Div pos)
						    
intoAdaptive' pos (Just (Ref x y))  = let row = rowNr x
				          col = colNr x
					  ref = (row,col)
			   	      in newMod (return 0) 
						>>= return.(CR ref) 
						>>= return.(Cell pos)

calc pos [] = Nothing
calc pos (h:t) | pos == getPos h = Just $ eval h
	       | otherwise       = calc pos t 

eval ::  GnuExpr GnuCellRef 
      -> InIO Changeable Float
eval (Const _ x)        = return x
eval (Add   _ l)        = ap (return sum)      (sequence (map eval l))
eval (Mul   _ l)        = ap (return product)  (sequence (map eval l))
eval (Sub   _ l)        = ap (return subtract) (sequence (map eval l))
eval (Div   _ l)        = ap (return divide)   (sequence (map eval l))
eval (Cell  _ (CR _ f)) = readMod f



intoCell :: (InM m,Monad (m IO IORef),NewMod m IO IORef) =>
		[((Int,Int),(IOMod (GnuExpr GnuCellRef),IOMod Float))]
		-> [GnuExpr GnuCellRef]
		-> m IO IORef [((Int,Int),(IOMod (GnuExpr GnuCellRef), IOMod Float))]

intoCell l [] = 
	return l
intoCell l (c@(Const pos value):cs) = do
	a@(pos,(expr,value)) <- newCell c 
	intoCell (a:l) cs
intoCell l (c@(Cell pos (CR ref _)):cs) = 
	case (lookup ref l) of Nothing -> intoCell l (cs++[c])
			       Just (expr, x)  -> do a <- newCell (Cell pos (CR ref x))
						     intoCell (a:l) cs
intoCell l (c:cs) =
	let pos     = getPos c
	    exprs   = getOperands c
	    visited = map fst l
	    exprsPos = getRefsPos exprs
	in if all (`elem` visited) exprsPos
	   then let l' = intoCell' l exprs
		in case c of 
			Add _ _ -> do a  <- newCell (Add pos l')
				      intoCell (a:l) cs
			Mul _ _ -> do a  <- newCell (Mul pos l') 
		   		      intoCell (a:l) cs 
			Div _ _ -> do a  <- newCell (Div pos l') 
		   		      intoCell (a:l) cs
			Sub _ _ -> do a  <- newCell (Sub pos l') 
		   		      intoCell (a:l) cs 
	   else intoCell l (cs++[c])



intoCell' v []                          = 
	[]
intoCell' v (a@(Const _ _):r)           = 
	a:(intoCell' v r)
intoCell' v (a@(Cell pos (CR ref _)):r) = 
	let value = (snd.fromJust) (lookup ref v)
	in (Cell pos (CR ref value)): intoCell' v r
intoCell' v (c:cs)                      =
	let pos  = getPos c
	    l    = getOperands c
	    l'   = intoCell' v l
	    expr = case c of
		Add _ _ -> Add pos l'
		Mul _ _ -> Mul pos l'
		Div _ _ -> Div pos l'
		Sub _ _ -> Sub pos l'
	in expr:(intoCell' v cs)



newCell ::  (NewMod n IO IORef) =>
            GnuExpr GnuCellRef
            -> n IO IORef ((Int, Int),
                           (Modifiable IO IORef (GnuExpr GnuCellRef), IOMod Float))
newCell  expr = do
    c <- newMod (return expr)
    v <- newMod $ readMod c >>= eval
    let pos@(row,col) = getPos expr
    newMod $ readMod v >>= inM . prAt (row+2) ((col)*35)     6 . show
    newMod $ readMod c >>= inM . prAt (row+2) ((col)*35+10) 10 . show
    return (pos,(c,v))

prAt l c w s = putStr (pos l c ++ replicate w ' ' ++ pos l c++s)
esc = ("\ESC["++)
pos l c = esc (show l++";"++show c++"H")
clear = pos 0 0 ++ esc "J"
cleareol = esc "K"


readPrompt :: Show a => a -> [Char] -> IO String
readPrompt c s = do prAt 20 c 0 (s++"> "++ cleareol)
                    s <- getLine
                    when (s == "quit") $ exitWith ExitSuccess
                    return s

msg s = prAt 19 0 0 (s ++ cleareol)

prompt :: (InM m) =>
             [((Int, Int), (IOMod (GnuExpr GnuCellRef), IOMod Float))] 
	  -> m IO IORef (IOMod (GnuExpr GnuCellRef), GnuExpr GnuCellRef)
prompt env = inM p where
   p = do s' <- readPrompt 0 "Cell"
	  s  <- return.read $ s'
          case lookup s env of
            Nothing -> do msg ("Cell " ++ show s ++ " not found")
                          p
            Just (c,v) -> do let r = do expr <- readPrompt 13 "Expr"
					let expr' = filter ((==[]).snd) $ parseExp env s expr
					case expr' of ((expr'',_):_) ->    msg "" >> return (c,expr'')
						      []             -> do msg "Syntax error"
									   r
                             r




main = run $ do inM $ putStr clear 
     		Gmr'Cells atrrs cels <- inM $ readGnumeric "simpleSpreadsheet.gnumeric"
		env   <- sequence $ intoAdaptive cels
		env'  <- intoCell [] env
		let loop = do (c,e) <- prompt env'
                   	      change c e
                   	      propagate
                   	      loop
     		loop

